#!/usr/bin/env python

# Copyright (C) 2007 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Validates a GTFS file.

For usage information run feedvalidator.py --help
"""

import bisect
import codecs
import datetime
from transitfeed.util import defaultdict
import optparse
import os
import os.path
import re
import socket
import sys
import time
import transitfeed
from transitfeed import TYPE_ERROR, TYPE_WARNING
from urllib2 import Request, urlopen, HTTPError, URLError
from transitfeed import util
import webbrowser

SVN_TAG_URL = 'http://googletransitdatafeed.googlecode.com/svn/tags/'


def MaybePluralizeWord(count, word):
  if count == 1:
    return word
  else:
    return word + 's'


def PrettyNumberWord(count, word):
  return '%d %s' % (count, MaybePluralizeWord(count, word))


def UnCamelCase(camel):
  return re.sub(r'([a-z])([A-Z])', r'\1 \2', camel)


def ProblemCountText(error_count, warning_count):
  results = []
  if error_count:
    results.append(PrettyNumberWord(error_count, 'error'))
  if warning_count:
    results.append(PrettyNumberWord(warning_count, 'warning'))

  return ' and '.join(results)


def CalendarSummary(schedule):
  today = datetime.date.today()
  summary_end_date = today + datetime.timedelta(days=60)
  start_date, end_date = schedule.GetDateRange()

  if not start_date or not end_date:
    return {}

  start_date_object = transitfeed.DateStringToDateObject(start_date)
  end_date_object = transitfeed.DateStringToDateObject(end_date)
  if not start_date_object or not end_date_object:
    return {}

  # Get the list of trips only during the period the feed is active.
  # As such we have to check if it starts in the future and/or if
  # if it ends in less than 60 days.
  date_trips_departures = schedule.GenerateDateTripsDeparturesList(
                              max(today, start_date_object),
                              min(summary_end_date, end_date_object))

  if not date_trips_departures:
    return {}

  # Check that the dates which will be shown in summary agree with these
  # calculations. Failure implies a bug which should be fixed. It isn't good
  # for users to discover assertion failures but means it will likely be fixed.
  assert start_date <= date_trips_departures[0][0].strftime("%Y%m%d")
  assert end_date >= date_trips_departures[-1][0].strftime("%Y%m%d")

  # Generate a map from int number of trips in a day to a list of date objects
  # with that many trips. The list of dates is sorted.
  trips_dates = defaultdict(lambda: [])
  trips = 0
  for date, day_trips, day_departures in date_trips_departures:
    trips += day_trips
    trips_dates[day_trips].append(date)
  mean_trips = trips / len(date_trips_departures)
  max_trips = max(trips_dates.keys())
  min_trips = min(trips_dates.keys())

  calendar_summary = {}
  calendar_summary['mean_trips'] = mean_trips
  calendar_summary['max_trips'] = max_trips
  calendar_summary['max_trips_dates'] = FormatDateList(trips_dates[max_trips])
  calendar_summary['min_trips'] = min_trips
  calendar_summary['min_trips_dates'] = FormatDateList(trips_dates[min_trips])
  calendar_summary['date_trips_departures'] = date_trips_departures
  calendar_summary['date_summary_range'] = "%s to %s" % (
      date_trips_departures[0][0].strftime("%a %b %d"),
      date_trips_departures[-1][0].strftime("%a %b %d"))

  return calendar_summary


def FormatDateList(dates):
  if not dates:
    return "0 service dates"

  formatted = [d.strftime("%a %b %d") for d in dates[0:3]]
  if len(dates) > 3:
    formatted.append("...")
  return "%s (%s)" % (PrettyNumberWord(len(dates), "service date"),
                      ", ".join(formatted))


def MaxVersion(versions):
  versions = filter(None, versions)
  versions.sort(lambda x,y: -cmp([int(item) for item in x.split('.')],
                                 [int(item) for item in y.split('.')]))
  if len(versions) > 0:
    return versions[0]


class CountingConsoleProblemAccumulator(transitfeed.SimpleProblemAccumulator):
  """Accumulate problems and count errors and warnings.

  Args:
    ignore_types: list of error type names that will be ignored. E.g.
                  ['ExpirationDate', 'UnusedStop']
  """

  def __init__(self, ignore_types=None):
    self._error_count = 0
    self._warning_count = 0
    self._ignore_types = ignore_types or set()

  def _Report(self, e):
    if e.__class__.__name__ in self._ignore_types:
      return
    transitfeed.SimpleProblemAccumulator._Report(self, e)
    if e.IsError():
      self._error_count += 1
    else:
      self._warning_count += 1

  def ErrorCount(self):
    return self._error_count

  def WarningCount(self):
    return self._warning_count

  def FormatCount(self):
    return ProblemCountText(self.ErrorCount(), self.WarningCount())

  def HasIssues(self):
    return self.ErrorCount() or self.WarningCount()


class BoundedProblemList(object):
  """A list of one type of ExceptionWithContext objects with bounded size."""
  def __init__(self, size_bound):
    self._count = 0
    self._exceptions = []
    self._size_bound = size_bound

  def Add(self, e):
    self._count += 1
    try:
      bisect.insort(self._exceptions, e)
    except TypeError:
      # The base class ExceptionWithContext raises this exception in __cmp__
      # to signal that an object is not comparable. Instead of keeping the most
      # significant issue keep the first reported.
      if self._count <= self._size_bound:
        self._exceptions.append(e)
    else:
      # self._exceptions is in order. Drop the least significant if the list is
      # now too long.
      if self._count > self._size_bound:
        del self._exceptions[-1]

  def _GetDroppedCount(self):
    return self._count - len(self._exceptions)

  def __repr__(self):
    return "<BoundedProblemList %s>" % repr(self._exceptions)

  count = property(lambda s: s._count)
  dropped_count = property(_GetDroppedCount)
  problems = property(lambda s: s._exceptions)


class LimitPerTypeProblemAccumulator(transitfeed.ProblemAccumulatorInterface):
  """Accumulate problems up to a maximum number per type.

  Args:
    limit_per_type: maximum number of errors and warnings to keep of each type
    ignore_types: list of error type names that will be ignored. E.g.
                  ['ExpirationDate', 'UnusedStop']
  """

  def __init__(self, limit_per_type, ignore_types=None):
    # {TYPE_WARNING: {"ClassName": BoundedProblemList()}}
    self._type_to_name_to_problist = {
      TYPE_WARNING: defaultdict(lambda: BoundedProblemList(limit_per_type)),
      TYPE_ERROR: defaultdict(lambda: BoundedProblemList(limit_per_type))
    }
    self._ignore_types = ignore_types or set()

  def HasIssues(self):
    return (self._type_to_name_to_problist[TYPE_ERROR] or
            self._type_to_name_to_problist[TYPE_WARNING])

  def _Report(self, e):
    if e.__class__.__name__ in self._ignore_types:
      return
    self._type_to_name_to_problist[e.GetType()][e.__class__.__name__].Add(e)

  def ErrorCount(self):
    error_sets = self._type_to_name_to_problist[TYPE_ERROR].values()
    return sum(map(lambda v: v.count, error_sets))

  def WarningCount(self):
    warning_sets = self._type_to_name_to_problist[TYPE_WARNING].values()
    return sum(map(lambda v: v.count, warning_sets))

  def ProblemList(self, problem_type, class_name):
    """Return the BoundedProblemList object for given type and class."""
    return self._type_to_name_to_problist[problem_type][class_name]

  def ProblemListMap(self, problem_type):
    """Return the map from class name to BoundedProblemList object."""
    return self._type_to_name_to_problist[problem_type]


class HTMLCountingProblemAccumulator(LimitPerTypeProblemAccumulator):
  def FormatType(self, f, level_name, class_problist):
    """Write the HTML dumping all problems of one type.

    Args:
      f: file object open for writing
      level_name: string such as "Error" or "Warning"
      class_problist: sequence of tuples (class name,
          BoundedProblemList object)
    """
    class_problist.sort()
    output = []
    for classname, problist in class_problist:
      output.append('<h4 class="issueHeader"><a name="%s%s">%s</a></h4><ul>\n' %
                    (level_name, classname, UnCamelCase(classname)))
      for e in problist.problems:
        self.FormatException(e, output)
      if problist.dropped_count:
        output.append('<li>and %d more of this type.' %
                      (problist.dropped_count))
      output.append('</ul>\n')
    f.write(''.join(output))

  def FormatTypeSummaryTable(self, level_name, name_to_problist):
    """Return an HTML table listing the number of problems by class name.

    Args:
      level_name: string such as "Error" or "Warning"
      name_to_problist: dict mapping class name to an BoundedProblemList object

    Returns:
      HTML in a string
    """
    output = []
    output.append('<table>')
    for classname in sorted(name_to_problist.keys()):
      problist = name_to_problist[classname]
      human_name = MaybePluralizeWord(problist.count, UnCamelCase(classname))
      output.append('<tr><td>%d</td><td><a href="#%s%s">%s</a></td></tr>\n' %
                    (problist.count, level_name, classname, human_name))
    output.append('</table>\n')
    return ''.join(output)

  def FormatException(self, e, output):
    """Append HTML version of e to list output."""
    d = e.GetDictToFormat()
    for k in ('file_name', 'feedname', 'column_name'):
      if k in d.keys():
        d[k] = '<code>%s</code>' % d[k]
    problem_text = e.FormatProblem(d).replace('\n', '<br>')
    output.append('<li>')
    output.append('<div class="problem">%s</div>' %
                  transitfeed.EncodeUnicode(problem_text))
    try:
      if hasattr(e, 'row_num'):
        line_str = 'line %d of ' % e.row_num
      else:
        line_str = ''
      output.append('in %s<code>%s</code><br>\n' %
                    (line_str, e.file_name))
      row = e.row
      headers = e.headers
      column_name = e.column_name
      table_header = ''  # HTML
      table_data = ''  # HTML
      for header, value in zip(headers, row):
        attributes = ''
        if header == column_name:
          attributes = ' class="problem"'
        table_header += '<th%s>%s</th>' % (attributes, header)
        table_data += '<td%s>%s</td>' % (attributes, value)
      # Make sure output is encoded into UTF-8
      output.append('<table class="dump"><tr>%s</tr>\n' %
                    transitfeed.EncodeUnicode(table_header))
      output.append('<tr>%s</tr></table>\n' %
                    transitfeed.EncodeUnicode(table_data))
    except AttributeError, e:
      pass  # Hope this was getting an attribute from e ;-)
    output.append('<br></li>\n')

  def FormatCount(self):
    return ProblemCountText(self.ErrorCount(), self.WarningCount())

  def CountTable(self):
    output = []
    output.append('<table class="count_outside">\n')
    output.append('<tr>')
    if self.ProblemListMap(TYPE_ERROR):
      output.append('<td><span class="fail">%s</span></td>' %
                    PrettyNumberWord(self.ErrorCount(), "error"))
    if self.ProblemListMap(TYPE_WARNING):
      output.append('<td><span class="fail">%s</span></td>' %
                    PrettyNumberWord(self.WarningCount(), "warning"))
    output.append('</tr>\n<tr>')
    if self.ProblemListMap(TYPE_ERROR):
      output.append('<td>\n')
      output.append(self.FormatTypeSummaryTable("Error",
                    self.ProblemListMap(TYPE_ERROR)))
      output.append('</td>\n')
    if self.ProblemListMap(TYPE_WARNING):
      output.append('<td>\n')
      output.append(self.FormatTypeSummaryTable("Warning",
                    self.ProblemListMap(TYPE_WARNING)))
      output.append('</td>\n')
    output.append('</table>')
    return ''.join(output)

  def WriteOutput(self, feed_location, f, schedule, other_problems, extension):
    """Write the html output to f."""
    if self.HasIssues():
      if self.ErrorCount() + self.WarningCount() == 1:
        summary = ('<span class="fail">Found this problem:</span>\n%s' %
                   self.CountTable())
      else:
        summary = ('<span class="fail">Found these problems:</span>\n%s' %
                   self.CountTable())
    else:
      summary = '<span class="pass">feed validated successfully</span>'
    if other_problems is not None:
      summary = ('<span class="fail">\n%s</span><br><br>' %
                 other_problems) + summary

    basename = os.path.basename(feed_location)
    feed_path = (feed_location[:feed_location.rfind(basename)], basename)

    agencies = ', '.join(['<a href="%s">%s</a>' % (a.agency_url, a.agency_name)
                          for a in schedule.GetAgencyList()])
    if not agencies:
      agencies = '?'

    dates = "No valid service dates found"
    (start, end) = schedule.GetDateRange()
    if start and end:
      def FormatDate(yyyymmdd):
        src_format = "%Y%m%d"
        dst_format = "%B %d, %Y"
        try:
          return time.strftime(dst_format,
                               time.strptime(yyyymmdd, src_format))
        except ValueError:
          return yyyymmdd

      formatted_start = FormatDate(start)
      formatted_end = FormatDate(end)
      dates = "%s to %s" % (formatted_start, formatted_end)

    calendar_summary = CalendarSummary(schedule)
    if calendar_summary:
      calendar_summary_html = """<br>
During the upcoming service dates %(date_summary_range)s:
<table>
<tr><th class="header">Average trips per date:</th><td class="header">%(mean_trips)s</td></tr>
<tr><th class="header">Most trips on a date:</th><td class="header">%(max_trips)s, on %(max_trips_dates)s</td></tr>
<tr><th class="header">Least trips on a date:</th><td class="header">%(min_trips)s, on %(min_trips_dates)s</td></tr>
</table>""" % calendar_summary
    else:
      calendar_summary_html = ""

    output_prefix = """
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
<title>FeedValidator: %(feed_file)s</title>
<style>
body {font-family: Georgia, serif; background-color: white}
.path {color: gray}
div.problem {max-width: 500px}
table.dump td,th {background-color: khaki; padding: 2px; font-family:monospace}
table.dump td.problem,th.problem {background-color: dc143c; color: white; padding: 2px; font-family:monospace}
table.count_outside td {vertical-align: top}
table.count_outside {border-spacing: 0px; }
table {border-spacing: 5px 0px; margin-top: 3px}
h3.issueHeader {padding-left: 0.5em}
h4.issueHeader {padding-left: 1em}
.pass {background-color: lightgreen}
.fail {background-color: yellow}
.pass, .fail {font-size: 16pt}
.header {background-color: white; font-family: Georgia, serif; padding: 0px}
th.header {text-align: right; font-weight: normal; color: gray}
.footer {font-size: 10pt}
</style>
</head>
<body>
GTFS validation results for feed:<br>
<code><span class="path">%(feed_dir)s</span><b>%(feed_file)s</b></code><br>
FeedValidator extension used: %(extension)s
<br><br>
<table>
<tr><th class="header">Agencies:</th><td class="header">%(agencies)s</td></tr>
<tr><th class="header">Routes:</th><td class="header">%(routes)s</td></tr>
<tr><th class="header">Stops:</th><td class="header">%(stops)s</td></tr>
<tr><th class="header">Trips:</th><td class="header">%(trips)s</td></tr>
<tr><th class="header">Shapes:</th><td class="header">%(shapes)s</td></tr>
<tr><th class="header">Effective:</th><td class="header">%(dates)s</td></tr>
</table>
%(calendar_summary)s
<br>
%(problem_summary)s
<br><br>
""" % { "feed_file": feed_path[1],
        "feed_dir": feed_path[0],
        "agencies": agencies,
        "routes": len(schedule.GetRouteList()),
        "stops": len(schedule.GetStopList()),
        "trips": len(schedule.GetTripList()),
        "shapes": len(schedule.GetShapeList()),
        "dates": dates,
        "problem_summary": summary,
        "calendar_summary": calendar_summary_html,
        "extension": extension}

# In output_suffix string
# time.strftime() returns a regular local time string (not a Unicode one) with
# default system encoding. And decode() will then convert this time string back
# into a Unicode string. We use decode() here because we don't want the operating
# system to do any system encoding (which may cause some problem if the string
# contains some non-English characters) for the string. Therefore we decode it
# back to its original Unicode code print.

    time_unicode = (time.strftime('%B %d, %Y at %I:%M %p %Z').
                    decode(sys.getfilesystemencoding()))
    output_suffix = """
<div class="footer">
Generated by <a href="http://code.google.com/p/googletransitdatafeed/wiki/FeedValidator">
FeedValidator</a> version %s on %s.
</div>
</body>
</html>""" % (transitfeed.__version__, time_unicode)

    f.write(transitfeed.EncodeUnicode(output_prefix))
    if self.ProblemListMap(TYPE_ERROR):
      f.write('<h3 class="issueHeader">Errors:</h3>')
      self.FormatType(f, "Error",
                      self.ProblemListMap(TYPE_ERROR).items())
    if self.ProblemListMap(TYPE_WARNING):
      f.write('<h3 class="issueHeader">Warnings:</h3>')
      self.FormatType(f, "Warning",
                      self.ProblemListMap(TYPE_WARNING).items())
    f.write(transitfeed.EncodeUnicode(output_suffix))


def RunValidationOutputFromOptions(feed, options):
  """Validate feed, output results per options and return an exit code."""
  if options.output.upper() == "CONSOLE":
    return RunValidationOutputToConsole(feed, options)
  else:
    return RunValidationOutputToFilename(feed, options, options.output)


def RunValidationOutputToFilename(feed, options, output_filename):
  """Validate feed, save HTML at output_filename and return an exit code."""
  try:
    output_file = open(output_filename, 'w')
    exit_code = RunValidationOutputToFile(feed, options, output_file)
    output_file.close()
  except IOError, e:
    print 'Error while writing %s: %s' % (output_filename, e)
    output_filename = None
    exit_code = 2

  if options.manual_entry and output_filename:
    webbrowser.open('file://%s' % os.path.abspath(output_filename))

  return exit_code


def RunValidationOutputToFile(feed, options, output_file):
  """Validate feed, write HTML to output_file and return an exit code."""
  accumulator = HTMLCountingProblemAccumulator(options.limit_per_type,
                                               options.error_types_ignore_list)
  problems = transitfeed.ProblemReporter(accumulator)
  schedule, exit_code, other_problems_string = RunValidation(feed, options,
                                                             problems)
  if isinstance(feed, basestring):
    feed_location = feed
  else:
    feed_location = getattr(feed, 'name', repr(feed))
  accumulator.WriteOutput(feed_location, output_file, schedule,
                          other_problems_string, options.extension)
  return exit_code


def RunValidationOutputToConsole(feed, options):
  """Validate feed, print reports and return an exit code."""
  accumulator = CountingConsoleProblemAccumulator(
      options.error_types_ignore_list)
  problems = transitfeed.ProblemReporter(accumulator)
  _, exit_code, _ = RunValidation(feed, options, problems)
  return exit_code


def RunValidation(feed, options, problems):
  """Validate feed, returning the loaded Schedule and exit code.

  Args:
    feed: GTFS file, either path of the file as a string or a file object
    options: options object returned by optparse
    problems: transitfeed.ProblemReporter instance

  Returns:
    a transitfeed.Schedule object, exit code and plain text string of other
    problems
    Exit code is 2 if an extension is provided but can't be loaded, 1 if
    problems are found and 0 if the Schedule is problem free.
    plain text string is '' if no other problems are found.
  """
  other_problems_string = CheckVersion(latest_version=options.latest_version)

  # TODO: Add tests for this flag in testfeedvalidator.py
  if options.extension:
    try:
      __import__(options.extension)
      extension_module = sys.modules[options.extension]
    except ImportError:
      # TODO: Document extensions in a wiki page, place link here
      print("Could not import extension %s! Please ensure it is a proper "
            "Python module." % options.extension)
      exit(2)
  else:
    extension_module = transitfeed

  gtfs_factory = extension_module.GetGtfsFactory()

  print 'validating %s' % feed
  print 'FeedValidator extension used: %s' % options.extension
  loader = gtfs_factory.Loader(feed, problems=problems, extra_validation=False,
                               memory_db=options.memory_db,
                               check_duplicate_trips=\
                               options.check_duplicate_trips,
                               gtfs_factory=gtfs_factory)
  schedule = loader.Load()
  # Start validation: children are already validated by the loader.
  schedule.Validate(service_gap_interval=options.service_gap_interval,
                    validate_children=False)
  
  if feed == 'IWantMyvalidation-crash.txt':
    # See test/testfeedvalidator.py
    raise Exception('For testing the feed validator crash handler.')

  if other_problems_string:
    print other_problems_string

  accumulator = problems.GetAccumulator()
  if accumulator.HasIssues():
    print 'ERROR: %s found' % accumulator.FormatCount()
    return schedule, 1, other_problems_string
  else:
    print 'feed validated successfully'
    return schedule, 0, other_problems_string


def CheckVersion(latest_version=''):
  """
  Check there is newer version of this project.

  Codes are based on http://www.voidspace.org.uk/python/articles/urllib2.shtml
  Already got permission from the copyright holder.
  """
  current_version = transitfeed.__version__
  if not latest_version:
    timeout = 20
    socket.setdefaulttimeout(timeout)
    request = Request(SVN_TAG_URL)

    try:
      response = urlopen(request)
      content = response.read()
      versions = re.findall(r'>transitfeed-([\d\.]+)\/<\/a>', content)
      latest_version = MaxVersion(versions)

    except HTTPError, e:
      return('The server couldn\'t fulfill the request. Error code: %s.'
             % e.code)
    except URLError, e:
      return('We failed to reach transitfeed server. Reason: %s.' % e.reason)

  if not latest_version:
    return('We had trouble parsing the contents of %s.' % SVN_TAG_URL)

  newest_version = MaxVersion([latest_version, current_version])
  if current_version != newest_version:
    return('A new version %s of transitfeed is available. Please visit '
           'http://code.google.com/p/googletransitdatafeed and download.'
           % newest_version)


def main():
  (feed, options) = ParseCommandLineArguments()
  return RunValidationFromOptions(feed, options)


def ParseCommandLineArguments():
  usage = \
'''%prog [options] [<input GTFS.zip>]

Validates GTFS file (or directory) <input GTFS.zip> and writes a HTML
report of the results to validation-results.html.

If <input GTFS.zip> is ommited the filename is read from the console. Dragging
a file into the console may enter the filename.

For more information see
http://code.google.com/p/googletransitdatafeed/wiki/FeedValidator
'''

  parser = util.OptionParserLongError(
      usage=usage, version='%prog '+transitfeed.__version__)
  parser.add_option('-n', '--noprompt', action='store_false',
                    dest='manual_entry',
                    help='do not prompt for feed location or load output in '
                    'browser')
  parser.add_option('-o', '--output', dest='output', metavar='FILE',
                    help='write html output to FILE or --output=CONSOLE to '
                    'print all errors and warnings to the command console')
  parser.add_option('-p', '--performance', action='store_true',
                    dest='performance',
                    help='output memory and time performance (Availability: '
                    'Unix')
  parser.add_option('-m', '--memory_db', dest='memory_db',  action='store_true',
                    help='Use in-memory sqlite db instead of a temporary file. '
                         'It is faster but uses more RAM.')
  parser.add_option('-d', '--duplicate_trip_check',
                    dest='check_duplicate_trips', action='store_true',
                    help='Check for duplicate trips which go through the same '
                    'stops with same service and start times')
  parser.add_option('-l', '--limit_per_type',
                    dest='limit_per_type', action='store', type='int',
                    help='Maximum number of errors and warnings to keep of '
                    'each type')
  parser.add_option('--latest_version', dest='latest_version',
                    action='store',
                    help='a version number such as 1.2.1 or None to get the '
                    'latest version from code.google.com. Output a warning if '
                    'transitfeed.py is older than this version.')
  parser.add_option('--service_gap_interval',
                    dest='service_gap_interval',
                    action='store',
                    type='int',
                    help='the number of consecutive days to search for with no '
                    'scheduled service. For each interval with no service '
                    'having this number of days or more a warning will be '
                    'issued')
  parser.add_option('--extension',
                    dest='extension',
                    help='the name of the Python module that containts a GTFS '
                    'extension that is to be loaded and used while validating '
                    'the specified feed.')
  parser.add_option('--error_types_ignore_list',
                    dest='error_types_ignore_list',
                    help='a comma-separated list of error and warning type '
                    'names to be ignored during validation (e.g. '
                    '"ExpirationDate,UnusedStop"). Bad error type names will '
                    'be silently ignored!')

  parser.set_defaults(manual_entry=True, output='validation-results.html',
                      memory_db=False, check_duplicate_trips=False,
                      limit_per_type=5, latest_version='',
                      service_gap_interval=13)
  (options, args) = parser.parse_args()

  if not len(args) == 1:
    if options.manual_entry:
      feed = raw_input('Enter Feed Location: ')
    else:
      parser.error('You must provide the path of a single feed')
  else:
    feed = args[0]
  feed = feed.strip('"')

  # transform options.error_types_ignore_list into a valid list
  if options.error_types_ignore_list:
    options.error_types_ignore_list = options.error_types_ignore_list.split(',')
  else:
    options.error_types_ignore_list = None

  return (feed, options)


def RunValidationFromOptions(feed, options):
  """Validate feed, run in profiler if in options, and return an exit code."""
  if options.performance:
    return ProfileRunValidationOutputFromOptions(feed, options)
  else:
    return RunValidationOutputFromOptions(feed, options)


def ProfileRunValidationOutputFromOptions(feed, options):
  """Run RunValidationOutputFromOptions, print profile and return exit code."""
  import cProfile
  import pstats
  # runctx will modify a dict, but not locals(). We need a way to get rv back.
  locals_for_exec = locals()
  cProfile.runctx('rv = RunValidationOutputFromOptions(feed, options)',
                  globals(), locals_for_exec, 'validate-stats')

  # Only available on Unix, http://docs.python.org/lib/module-resource.html
  import resource
  print "Time: %d seconds" % (
      resource.getrusage(resource.RUSAGE_SELF).ru_utime +
      resource.getrusage(resource.RUSAGE_SELF).ru_stime)

  # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/286222
  # http://aspn.activestate.com/ASPN/Cookbook/ "The recipes are freely
  # available for review and use."
  def _VmB(VmKey):
    """Return size from proc status in bytes."""
    _proc_status = '/proc/%d/status' % os.getpid()
    _scale = {'kB': 1024.0, 'mB': 1024.0*1024.0,
              'KB': 1024.0, 'MB': 1024.0*1024.0}

     # get pseudo file  /proc/<pid>/status
    try:
        t = open(_proc_status)
        v = t.read()
        t.close()
    except:
        raise Exception("no proc file %s" % _proc_status)
        return 0  # non-Linux?
     # get VmKey line e.g. 'VmRSS:  9999  kB\n ...'
    try:
        i = v.index(VmKey)
        v = v[i:].split(None, 3)  # whitespace
    except:
        return 0 # v is empty

    if len(v) < 3:
        raise Exception("%s" % v)
        return 0  # invalid format?
     # convert Vm value to bytes
    return int(float(v[1]) * _scale[v[2]])

  # I ran this on over a hundred GTFS files, comparing VmSize to VmRSS
  # (resident set size). The difference was always under 2% or 3MB.
  print "Virtual Memory Size: %d bytes" % _VmB('VmSize:')

  # Output report of where CPU time was spent.
  p = pstats.Stats('validate-stats')
  p.strip_dirs()
  p.sort_stats('cumulative').print_stats(30)
  p.sort_stats('cumulative').print_callers(30)
  return locals_for_exec['rv']


if __name__ == '__main__':
  util.RunWithCrashHandler(main)




import os
from os.path import join as pathjoin, abspath, dirname, exists as pathexists


PACKAGE_DATA_ROOT = pathjoin(abspath(dirname(__file__)), 'data')
try:
    from django.conf import settings
    APPLICATION_DATA_ROOT = settings.DATA_ROOT
except (ImportError, AttributeError):
    APPLICATION_DATA_ROOT = None
if not APPLICATION_DATA_ROOT:
    APPLICATION_DATA_ROOT = pathjoin(abspath(os.getcwd()), 'data')

def get_pkg_data(path):
    return pathjoin(PACKAGE_DATA_ROOT, path)

def get_app_data(path, ensure_root_exists=True):
    fpath = pathjoin(APPLICATION_DATA_ROOT, path)
    if ensure_root_exists:
        if fpath.endswith('/'):
            root = fpath
        else:
            root = dirname(fpath)
        if not pathexists(root):
            os.makedirs(root)
    return fpath

def make_app_dir(path):
    dirpath = pathjoin(APPLICATION_DATA_ROOT, path)
    if not pathexists(dirpath):
        os.makedirs(dirpath)
    return dirpath





import re
import string

NORMALISATIONS = reversed(sorted([
    ('road', 'rd'),
    ('avenue', 'ave'),
    ('street', 'st'),
    ('place', 'pl'),
    ('estate', 'est'),
    ('pk', 'park'),
    ('sq', 'square'),
    ('stn', 'station'),
    ('cres', 'crescent'),
    ('cen', 'centre'),
    ('sth', 'south'),
    ('nth', 'north'),
    ('pde', 'parade'),
    ('drv', 'drive'),
    ('la', 'lane'), #Gray's Lane
    ('jct', 'junction'),
    ('fm ctr', 'farm'), #Dale farm
    ('mt', 'mount'),
    ('bus stop', ''),
    ('boucher rd, charles hurst', 'hursts, boucher crescent'),
    ("k'breda", 'knockbreda'),
    ('cemy', 'cemetery'),
    ('cemetry', 'cemetery'),
    ('st.james', 'st. james'),
    ('st.teresa', 'st. teresa'),
    ('st.theresa', 'st. teresa'),
    ('hospitals', 'hospital'), #Royal Victoria
    ('trossachs dr', 'trossachs'),
    ('kingsway park', 'kingsway'),
    ('george best', 'city'),
    ('sydenham by-pass', 'belfast'),
    ('skipper street', ''),
    ('hydebank r/bout', 'hydebank'),
    ('ligoniel, mountainhill', 'ligoniel'),
    ('mossley, glade', 'mossley(rail station)'),
    ('newton park, colby park', 'newton park(colby)'),
    ('donegall', 'donegal'),
    ('queens square albert clock', 'queens square'),
]))
rx_replace = re.compile(r'[^-\w\s,()/]').sub
rx_hyphenate = re.compile(r'[-\s,()/]+').sub
rx_normalise = [
    (re.compile('\\b'+patt+'\\b'), repl) for patt, repl in NORMALISATIONS
]

def slugify(s):
    s = str(s).strip().lower()
    if not s:
        return ''
    for patt, repl in rx_normalise:
        s = patt.sub(repl, s)
    return rx_hyphenate('-', rx_replace('', s).strip()).strip('-')

direction2name = {0: 'Outbound', 1: 'Inbound'}
direction2code = {'Outbound': 0, 'Inbound': 1, 'outbound': 0, 'inbound': 1}


class MultiReplace:

    def __init__(self, repl_dict):
        # "compile" replacement dictionary

        # assume char to char mapping
        charmap = map(chr, range(256))
        for k, v in repl_dict.items():
            if len(k) != 1 or len(v) != 1:
                self.charmap = None
                break
            charmap[ord(k)] = v
        else:
            self.charmap = string.join(charmap, "")
            return

        # string to string mapping; use a regular expression
        keys = repl_dict.keys()
        keys.sort() # lexical order
        keys.reverse() # use longest match first
        pattern = string.join(map(re.escape, keys), "|")
        self.pattern = re.compile(pattern)
        self.dict = repl_dict

    def replace(self, str):
        # apply replacement dictionary to string
        if self.charmap:
            return string.translate(str, self.charmap)
        def repl(match, get=self.dict.get):
            item = match.group(0)
            return get(item, item)
        return self.pattern.sub(repl, str)

daycodes = 'M T W TH F S SU'.split()
daynames = 'Monday Tuesday Wednesday Thursday Friday Saturday Sunday'.split()
daycode_to_name = dict(zip(daycodes, daynames))
replace_names = MultiReplace(dict(zip(daynames, daycodes))).replace
sdaycodes = ''.join(daycodes)
daycode_atomic_weights = [2**i for i in range(7, 0, -1)]

#TODO: use itertools
def xcombinations(items, n):
    if n==0:
        yield []
    else:
        for i in xrange(len(items)):
            #yield items[:i]
            for cc in xcombinations(items[i+1:],n-1):
                yield [items[i]]+cc


def idaycode_combos():
    for i in xrange(1, 8):
        for rng in xcombinations(daycodes, i):
            names = [daycode_to_name[code] for code in rng]
            id = ''.join(rng)
            # the weight gives precedence to days that are earlier in the 
            # week, so that, for a given set of disjoint time periods,
            # earlier periods can be displayed before later periods
            # eg. Monday-Friday, Saturday, Sunday
            # eg. "Monday, Wednesday, Friday" and "Tuesday, Thursday, Saturday"
            # it is up to the application to ensure that time periods for
            # any given route *are* disjoint
            weight = sum(2**(7-daycodes.index(s)) for s in rng)
            if i == 1:
                yield id, (weight, names[0])
            else:
                if len(rng) > 2 and id in sdaycodes:
                    # consecutive days, eg. Monday to Thursday
                    displayname = '%s to %s' % (names[0], names[-1])
                    # want to yield both options, eg. M-W and MTW
                    # this implies weight is not unique in resulting dataset
                    yield id, (weight, displayname)
                    yield '%s-%s' % (rng[0], rng[-1]), (weight, displayname)
                else:
                    names[-2:] = ['%s and %s' % (names[-2], names[-1])]
                    displayname = ', '.join(names)
                    yield id, (weight, displayname)

timeframe_combinations = dict(idaycode_combos())

def get_time_period_name_and_weight(s):
    """
    >>> get = get_time_period_name_and_weight
    >>> name1, weight1 = get('Monday - Friday')
    >>> name2, weight2 = get('M-F')
    >>> name3, weight3 = get('MTWTHF')
    >>> weight1 == weight2 == weight3
    True
    >>> name1 == name2 == name3
    True
    >>> name1
    'Monday to Friday'
    >>> get('WTHSU')
    ('Wednesday, Thursday and Sunday', 50)

    Days can be retrieved from weight::

    >>> assert 50 | (2**7) != 50
    >>> assert 50 | (2**6) != 50
    >>> assert 50 | (2**5) == 50 # Wednesday
    >>> assert 50 | (2**4) == 50 # Thursday
    >>> assert 50 | (2**3) != 50
    >>> assert 50 | (2**2) != 50
    >>> assert 50 | (2**1) == 50 # Sunday

    Compare weights::

    >>> weight1 = get('MWF')[1]
    >>> weight2 = get('TTHS')[1]
    >>> weight1 > weight2
    True
    >>> weight1 = get('M')[1]
    >>> weight2 = get('T-SU')[1]
    >>> weight1 > weight2
    True
    """
    weight, name = timeframe_combinations[''.join(replace_names(s).split()).upper()]
    return name, weight

def split_timeframe(tf):
    """
    >>> list(split_timeframe('M-F'))
    ['M', 'T', 'W', 'TH', 'F']
    >>> list(split_timeframe('S'))
    ['S']
    >>> list(split_timeframe('Th-Su'))
    ['TH', 'F', 'S', 'SU']
    >>> list(split_timeframe('MFS'))
    ['M', 'F', 'S']
    """
    name, weight = get_time_period_name_and_weight(tf)
    for i, (w, daycode) in enumerate(zip(daycode_atomic_weights, daycodes)):
        if weight & w:
            yield i, daycode

if __name__ == '__main__':
    import doctest
    doctest.testmod()




#!/usr/bin/env python

import os
import sys
import re
import csv

import requests
from geohash import encode as geohash

from gtfsni import get_pkg_data, get_app_data
from gtfsni.utils import slugify, direction2name

FIXES_FILE = get_pkg_data('translink/metro_stops_fixes.csv')

assert os.path.exists(FIXES_FILE), "missing data file - %s" % FIXES_FILE

ENDPOINT = 'https://api.scraperwiki.com/api/1.0/datastore/sqlite'
REFERENCE_SCHEMA = [
    'name', 'road', 'lat', 'lng', 'direction'
]
TIMETABLE_SCHEMA = [
    'stop_name', 'route_direction',
]
STOPS_SCHEMA = [
    'stop_lat', 'stop_lon', 'stop_name', 'road', 'direction', 'geohash'
]
FIXES_SCHEMA = [
    'stop_name', 'route_direction', 'stop_lat', 'stop_lon', 'road',
]
INBOUND_REFERENCE = {}
OUTBOUND_REFERENCE = {}
REFERENCE_STOPS = {}
REFERENCE_STOPS['Inbound'] = REFERENCE_STOPS[1] = INBOUND_REFERENCE
REFERENCE_STOPS['Outbound'] = REFERENCE_STOPS[0] = OUTBOUND_REFERENCE
TIMETABLE_STOPS = None
OPENTRANSLINK_STOPS = {}
FIXES = {}

def ReferenceReader(fd):
    return csv.DictReader(fd, REFERENCE_SCHEMA)

def TimetableReader(fd):
    return csv.DictReader(fd, TIMETABLE_SCHEMA)

def FixesReader(fd):
    return csv.DictReader(fd, FIXES_SCHEMA)

def StopsWriter(fd):
    return csv.DictWriter(fd, STOPS_SCHEMA)

def StopsHeader():
    return dict(zip(STOPS_SCHEMA, STOPS_SCHEMA))

def StopsRow():
    return dict((field, None) for field in STOPS_SCHEMA)

def load_reference_stops():
    sql = "SELECT DISTINCT name, road, lat, lng, direction from `swdata`"
    params = {
        'name': 'translink_metro_stops',
        'format': 'csv',
        'query': sql,
    }
    response = requests.get(ENDPOINT, params=params)
    reader = ReferenceReader(response.text.splitlines())
    reader.next()
    for line in reader:
        # generate variations - all these will give the same lat/lon
        possibilities = [
            slugify(line['name']),
            slugify(line['name'] + ' ' + line['road']),
            slugify(line['road']),
        ]
        try:
            stops = [REFERENCE_STOPS[line['direction']]]
        except KeyError:
            # no direction given - add the stop to both reference dicts
            stops = [REFERENCE_STOPS[0], REFERENCE_STOPS[1]]
        for reference_dict in stops:
            for possibility in possibilities:
                reference_dict[possibility] = (
                    float(line['lat']), float(line['lng']), line['road'],
                )

def load_timetable_stops():
    global TIMETABLE_STOPS
    sql = (
        "SELECT DISTINCT stop_name, route_direction from `swdata`"
        " WHERE operator_id='METRO'"
    )
    params = {
        'name': 'translink_ni_timetables',
        'format': 'csv',
        'query': sql,
    }
    response = requests.get(ENDPOINT, params=params)
    if response.status_code != 200:
        sys.exit("GET '%s' failed. Error: %s" % (response.url, response.error))
    reader = TimetableReader(response.text.splitlines())
    reader.next()
    TIMETABLE_STOPS = [(row['stop_name'], row['route_direction']) for row in reader]

def load_opentranslink_stops():
    with open(get_pkg_data('translink/opentranslink_stops.csv')) as fd:
        reader = csv.reader(fd)
        for row in reader:
            key = slugify(row[0])
            val = (row[1], row[2],'')
            OPENTRANSLINK_STOPS[key] = val

def load_manual_fixes():
    with open(FIXES_FILE) as fd:
        reader = FixesReader(fd)
        reader.next()
        for row in reader:
            key = (row['stop_name'], int(row['route_direction']))
            val = (row['stop_lat'], row['stop_lon'], row['road'])
            FIXES[key] = val


def _find_info(stop, stops, top, tail):
    slug = slugify(stop)
    info = stops.get(slug, None)
    if not info and tail:
        slug = slugify(tail + ' ' + top)
        info = stops.get(slug, None)
    return info

def find_info(stop, direction):
    """
    Heuristic search for the named stop, first try the direction given, then
    the other direction (across the road will be better than nothing), then
    if the stop has more than one part (comma-separated), try the second part
    on its own.
    """
    stops = REFERENCE_STOPS[direction]
    other_stops = REFERENCE_STOPS[int(not direction)]
    top, comma, tail = stop.partition(',')
    info = _find_info(stop, stops, top, tail)
    if not info:
        info = _find_info(stop, other_stops, top, tail)
    if not info and tail:
        info = _find_info(tail, stops, None, None)
        if not info:
            info = _find_info(tail, other_stops, None, None)
    if not info:
        info = _find_info(stop, OPENTRANSLINK_STOPS, None, None)
    if info:
        return info

def main():
    """
    Load the stops and coordinate data as derived from the map view pages to
    the REFERENCE_STOPS dict. Load the stop names as derived from the timetable
    pages to TIMETABLE_STOPS list. Load the manually-added stop coordinates to
    FIXES dict.
    """
    load_reference_stops()
    load_timetable_stops()
    load_opentranslink_stops()
    load_manual_fixes()
    found = missing = 0
    with open(get_app_data('translink/metro_stops.csv'), 'wb') as fd:
        writer = StopsWriter(fd)
        writer.writerow(StopsHeader())
        for name, direction in TIMETABLE_STOPS:
            direction = int(direction) # 0 or 1
            row = StopsRow()
            row['stop_name'] = name
            row['direction'] = direction2name[direction]
            try:
                # has it been added manually?
                info = FIXES[(name, direction)]
            except KeyError:
                # try to find from the google map-scraped coords
                info = find_info(name, direction)
            if info:
                # info is a 3-tuple
                x, y = row['stop_lat'], row['stop_lon'] = info[:2]
                row['road'] = info[2]
                row['geohash'] = geohash(float(x), float(y))
                found += 1
            else:
                print "MISSING: ", name, direction
                missing += 1
            writer.writerow(row)
    print 'found: ', found, ', missing: %s' % missing

if __name__ == '__main__':
    main()




#!/usr/bin/python2.5

# Copyright (C) 2007 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An example application that uses the transitfeed module.

You must provide a Google Maps API key.
"""


import BaseHTTPServer, sys, urlparse
import bisect
from gtfsscheduleviewer.marey_graph import MareyGraph
import gtfsscheduleviewer
import mimetypes
import os.path
import re
import signal
import simplejson
import socket
import time
import transitfeed
from transitfeed import util
import urllib


# By default Windows kills Python with Ctrl+Break. Instead make Ctrl+Break
# raise a KeyboardInterrupt.
if hasattr(signal, 'SIGBREAK'):
  signal.signal(signal.SIGBREAK, signal.default_int_handler)


mimetypes.add_type('text/plain', '.vbs')


class ResultEncoder(simplejson.JSONEncoder):
  def default(self, obj):
    try:
      iterable = iter(obj)
    except TypeError:
      pass
    else:
      return list(iterable)
    return simplejson.JSONEncoder.default(self, obj)

# Code taken from
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/425210/index_txt
# An alternate approach is shown at
# http://mail.python.org/pipermail/python-list/2003-July/212751.html
# but it requires multiple threads. A sqlite object can only be used from one
# thread.
class StoppableHTTPServer(BaseHTTPServer.HTTPServer):
  def server_bind(self):
    BaseHTTPServer.HTTPServer.server_bind(self)
    self.socket.settimeout(1)
    self._run = True

  def get_request(self):
    while self._run:
      try:
        sock, addr = self.socket.accept()
        sock.settimeout(None)
        return (sock, addr)
      except socket.timeout:
        pass

  def stop(self):
    self._run = False

  def serve(self):
    while self._run:
      self.handle_request()


def StopToTuple(stop):
  """Return tuple as expected by javascript function addStopMarkerFromList"""
  return (stop.stop_id, stop.stop_name, float(stop.stop_lat),
          float(stop.stop_lon), stop.location_type)


class ScheduleRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
  def do_GET(self):
    scheme, host, path, x, params, fragment = urlparse.urlparse(self.path)
    parsed_params = {}
    for k in params.split('&'):
      k = urllib.unquote(k)
      if '=' in k:
        k, v = k.split('=', 1)
        parsed_params[k] = unicode(v, 'utf8')
      else:
        parsed_params[k] = ''

    if path == '/':
      return self.handle_GET_home()

    m = re.match(r'/json/([a-z]{1,64})', path)
    if m:
      handler_name = 'handle_json_GET_%s' % m.group(1)
      handler = getattr(self, handler_name, None)
      if callable(handler):
        return self.handle_json_wrapper_GET(handler, parsed_params)

    # Restrict allowable file names to prevent relative path attacks etc
    m = re.match(r'/file/([a-z0-9_-]{1,64}\.?[a-z0-9_-]{1,64})$', path)
    if m and m.group(1):
      try:
        f, mime_type = self.OpenFile(m.group(1))
        return self.handle_static_file_GET(f, mime_type)
      except IOError, e:
        print "Error: unable to open %s" % m.group(1)
        # Ignore and treat as 404

    m = re.match(r'/([a-z]{1,64})', path)
    if m:
      handler_name = 'handle_GET_%s' % m.group(1)
      handler = getattr(self, handler_name, None)
      if callable(handler):
        return handler(parsed_params)

    return self.handle_GET_default(parsed_params, path)

  def OpenFile(self, filename):
    """Try to open filename in the static files directory of this server.
    Return a tuple (file object, string mime_type) or raise an exception."""
    (mime_type, encoding) = mimetypes.guess_type(filename)
    assert mime_type
    # A crude guess of when we should use binary mode. Without it non-unix
    # platforms may corrupt binary files.
    if mime_type.startswith('text/'):
      mode = 'r'
    else:
      mode = 'rb'
    return open(os.path.join(self.server.file_dir, filename), mode), mime_type

  def handle_GET_default(self, parsed_params, path):
    self.send_error(404)

  def handle_static_file_GET(self, fh, mime_type):
    content = fh.read()
    self.send_response(200)
    self.send_header('Content-Type', mime_type)
    self.send_header('Content-Length', str(len(content)))
    self.end_headers()
    self.wfile.write(content)

  def AllowEditMode(self):
    return False

  def handle_GET_home(self):
    schedule = self.server.schedule
    (min_lat, min_lon, max_lat, max_lon) = schedule.GetStopBoundingBox()
    forbid_editing = ('true', 'false')[self.AllowEditMode()]

    agency = ', '.join(a.agency_name for a in schedule.GetAgencyList()).encode('utf-8')

    key = self.server.key
    host = self.server.host

    # A very simple template system. For a fixed set of values replace [xxx]
    # with the value of local variable xxx
    f, _ = self.OpenFile('index.html')
    content = f.read()
    for v in ('agency', 'min_lat', 'min_lon', 'max_lat', 'max_lon', 'key',
              'host', 'forbid_editing'):
      content = content.replace('[%s]' % v, str(locals()[v]))

    self.send_response(200)
    self.send_header('Content-Type', 'text/html')
    self.send_header('Content-Length', str(len(content)))
    self.end_headers()
    self.wfile.write(content)

  def handle_json_GET_routepatterns(self, params):
    """Given a route_id generate a list of patterns of the route. For each
    pattern include some basic information and a few sample trips."""
    schedule = self.server.schedule
    route = schedule.GetRoute(params.get('route', None))
    if not route:
      self.send_error(404)
      return
    time = int(params.get('time', 0))
    date = params.get('date', "")
    sample_size = 3  # For each pattern return the start time for this many trips

    pattern_id_trip_dict = route.GetPatternIdTripDict()
    patterns = []

    for pattern_id, trips in pattern_id_trip_dict.items():
      time_stops = trips[0].GetTimeStops()
      if not time_stops:
        continue
      has_non_zero_trip_type = False;

      # Iterating over a copy so we can remove from trips inside the loop
      trips_with_service = []
      for trip in trips:
        service_id = trip.service_id
        service_period = schedule.GetServicePeriod(service_id)

        if date and not service_period.IsActiveOn(date):
          continue
        trips_with_service.append(trip)

        if trip['trip_type'] and trip['trip_type'] != '0':
          has_non_zero_trip_type = True

      # We're only interested in the trips that do run on the specified date
      trips = trips_with_service

      name = u'%s to %s, %d stops' % (time_stops[0][2].stop_name, time_stops[-1][2].stop_name, len(time_stops))
      transitfeed.SortListOfTripByTime(trips)

      num_trips = len(trips)
      if num_trips <= sample_size:
        start_sample_index = 0
        num_after_sample = 0
      else:
        # Will return sample_size trips that start after the 'time' param.

        # Linear search because I couldn't find a built-in way to do a binary
        # search with a custom key.
        start_sample_index = len(trips)
        for i, trip in enumerate(trips):
          if trip.GetStartTime() >= time:
            start_sample_index = i
            break

        num_after_sample = num_trips - (start_sample_index + sample_size)
        if num_after_sample < 0:
          # Less than sample_size trips start after 'time' so return all the
          # last sample_size trips.
          num_after_sample = 0
          start_sample_index = num_trips - sample_size

      sample = []
      for t in trips[start_sample_index:start_sample_index + sample_size]:
        sample.append( (t.GetStartTime(), t.trip_id) )

      patterns.append((name, pattern_id, start_sample_index, sample,
                       num_after_sample, (0,1)[has_non_zero_trip_type]))

    patterns.sort()
    return patterns

  def handle_json_wrapper_GET(self, handler, parsed_params):
    """Call handler and output the return value in JSON."""
    schedule = self.server.schedule
    result = handler(parsed_params)
    content = ResultEncoder().encode(result)
    self.send_response(200)
    self.send_header('Content-Type', 'text/plain')
    self.send_header('Content-Length', str(len(content)))
    self.end_headers()
    self.wfile.write(content)

  def handle_json_GET_routes(self, params):
    """Return a list of all routes."""
    schedule = self.server.schedule
    result = []
    for r in schedule.GetRouteList():
      result.append( (r.route_id, r.route_short_name, r.route_long_name) )
    result.sort(key = lambda x: x[1:3])
    return result

  def handle_json_GET_routerow(self, params):
    schedule = self.server.schedule
    route = schedule.GetRoute(params.get('route', None))
    return [transitfeed.Route._FIELD_NAMES, route.GetFieldValuesTuple()]

  def handle_json_GET_triprows(self, params):
    """Return a list of rows from the feed file that are related to this
    trip."""
    schedule = self.server.schedule
    try:
      trip = schedule.GetTrip(params.get('trip', None))
    except KeyError:
      # if a non-existent trip is searched for, the return nothing
      return
    route = schedule.GetRoute(trip.route_id)
    trip_row = dict(trip.iteritems())
    route_row = dict(route.iteritems())
    return [['trips.txt', trip_row], ['routes.txt', route_row]]

  def handle_json_GET_tripstoptimes(self, params):
    schedule = self.server.schedule
    try:
      trip = schedule.GetTrip(params.get('trip'))
    except KeyError:
       # if a non-existent trip is searched for, the return nothing
      return
    time_stops = trip.GetTimeStops()
    stops = []
    arrival_times = []
    departure_times = []
    for arr,dep,stop in time_stops:
      stops.append(StopToTuple(stop))
      arrival_times.append(arr)
      departure_times.append(dep)
    return [stops, arrival_times, departure_times]

  def handle_json_GET_tripshape(self, params):
    schedule = self.server.schedule
    try:
      trip = schedule.GetTrip(params.get('trip'))
    except KeyError:
       # if a non-existent trip is searched for, the return nothing
      return
    points = []
    if trip.shape_id:
      shape = schedule.GetShape(trip.shape_id)
      for (lat, lon, dist) in shape.points:
        points.append((lat, lon))
    else:
      time_stops = trip.GetTimeStops()
      for arr,dep,stop in time_stops:
        points.append((stop.stop_lat, stop.stop_lon))
    return points

  def handle_json_GET_neareststops(self, params):
    """Return a list of the nearest 'limit' stops to 'lat', 'lon'"""
    schedule = self.server.schedule
    lat = float(params.get('lat'))
    lon = float(params.get('lon'))
    limit = int(params.get('limit'))
    stops = schedule.GetNearestStops(lat=lat, lon=lon, n=limit)
    return [StopToTuple(s) for s in stops]

  def handle_json_GET_boundboxstops(self, params):
    """Return a list of up to 'limit' stops within bounding box with 'n','e'
    and 's','w' in the NE and SW corners. Does not handle boxes crossing
    longitude line 180."""
    schedule = self.server.schedule
    n = float(params.get('n'))
    e = float(params.get('e'))
    s = float(params.get('s'))
    w = float(params.get('w'))
    limit = int(params.get('limit'))
    stops = schedule.GetStopsInBoundingBox(north=n, east=e, south=s, west=w, n=limit)
    return [StopToTuple(s) for s in stops]

  def handle_json_GET_stopsearch(self, params):
    schedule = self.server.schedule
    query = params.get('q', None).lower()
    matches = []
    for s in schedule.GetStopList():
      if s.stop_id.lower().find(query) != -1 or s.stop_name.lower().find(query) != -1:
        matches.append(StopToTuple(s))
    return matches

  def handle_json_GET_stoptrips(self, params):
    """Given a stop_id and time in seconds since midnight return the next
    trips to visit the stop."""
    schedule = self.server.schedule
    stop = schedule.GetStop(params.get('stop', None))
    time = int(params.get('time', 0))
    date = params.get('date', "")

    time_trips = stop.GetStopTimeTrips(schedule)
    time_trips.sort()  # OPT: use bisect.insort to make this O(N*ln(N)) -> O(N)
    # Keep the first 5 after param 'time'.
    # Need make a tuple to find correct bisect point
    time_trips = time_trips[bisect.bisect_left(time_trips, (time, 0)):]
    time_trips = time_trips[:5]
    # TODO: combine times for a route to show next 2 departure times
    result = []
    for time, (trip, index), tp in time_trips:
      service_id = trip.service_id
      service_period = schedule.GetServicePeriod(service_id)
      if date and not service_period.IsActiveOn(date):
        continue
      headsign = None
      # Find the most recent headsign from the StopTime objects
      for stoptime in trip.GetStopTimes()[index::-1]:
        if stoptime.stop_headsign:
          headsign = stoptime.stop_headsign
          break
      # If stop_headsign isn't found, look for a trip_headsign
      if not headsign:
        headsign = trip.trip_headsign
      route = schedule.GetRoute(trip.route_id)
      trip_name = ''
      if route.route_short_name:
        trip_name += route.route_short_name
      if route.route_long_name:
        if len(trip_name):
          trip_name += " - "
        trip_name += route.route_long_name
      if headsign:
        trip_name += " (Direction: %s)" % headsign

      result.append((time, (trip.trip_id, trip_name, trip.service_id), tp))
    return result

  def handle_GET_ttablegraph(self,params):
    """Draw a Marey graph in SVG for a pattern (collection of trips in a route
    that visit the same sequence of stops)."""
    schedule = self.server.schedule
    marey = MareyGraph()
    trip = schedule.GetTrip(params.get('trip', None))
    route = schedule.GetRoute(trip.route_id)
    height = int(params.get('height', 300))

    if not route:
      print 'no such route'
      self.send_error(404)
      return

    pattern_id_trip_dict = route.GetPatternIdTripDict()
    pattern_id = trip.pattern_id
    if pattern_id not in pattern_id_trip_dict:
      print 'no pattern %s found in %s' % (pattern_id, pattern_id_trip_dict.keys())
      self.send_error(404)
      return
    triplist = pattern_id_trip_dict[pattern_id]

    pattern_start_time = min((t.GetStartTime() for t in triplist))
    pattern_end_time = max((t.GetEndTime() for t in triplist))

    marey.SetSpan(pattern_start_time,pattern_end_time)
    marey.Draw(triplist[0].GetPattern(), triplist, height)

    content = marey.Draw()

    self.send_response(200)
    self.send_header('Content-Type', 'image/svg+xml')
    self.send_header('Content-Length', str(len(content)))
    self.end_headers()
    self.wfile.write(content)


def FindPy2ExeBase():
  """If this is running in py2exe return the install directory else return
  None"""
  # py2exe puts gtfsscheduleviewer in library.zip. For py2exe setup.py is
  # configured to put the data next to library.zip.
  windows_ending = gtfsscheduleviewer.__file__.find('\\library.zip\\')
  if windows_ending != -1:
    return transitfeed.__file__[:windows_ending]
  else:
    return None


def FindDefaultFileDir():
  """Return the path of the directory containing the static files. By default
  the directory is called 'files'. The location depends on where setup.py put
  it."""
  base = FindPy2ExeBase()
  if base:
    return os.path.join(base, 'schedule_viewer_files')
  else:
    # For all other distributions 'files' is in the gtfsscheduleviewer
    # directory.
    base = os.path.dirname(gtfsscheduleviewer.__file__)  # Strip __init__.py
    return os.path.join(base, 'files')


def GetDefaultKeyFilePath():
  """In py2exe return absolute path of file in the base directory and in all
  other distributions return relative path 'key.txt'"""
  windows_base = FindPy2ExeBase()
  if windows_base:
    return os.path.join(windows_base, 'key.txt')
  else:
    return 'key.txt'


def main(RequestHandlerClass = ScheduleRequestHandler):
  usage = \
'''%prog [options] [<input GTFS.zip>]

Runs a webserver that lets you explore a <input GTFS.zip> in your browser.

If <input GTFS.zip> is omited the filename is read from the console. Dragging
a file into the console may enter the filename.

For more information see
http://code.google.com/p/googletransitdatafeed/wiki/ScheduleViewer
'''
  parser = util.OptionParserLongError(
      usage=usage, version='%prog '+transitfeed.__version__)
  parser.add_option('--feed_filename', '--feed', dest='feed_filename',
                    help='file name of feed to load')
  parser.add_option('--key', dest='key',
                    help='Google Maps API key or the name '
                    'of a text file that contains an API key')
  parser.add_option('--host', dest='host', help='Host name of Google Maps')
  parser.add_option('--port', dest='port', type='int',
                    help='port on which to listen')
  parser.add_option('--file_dir', dest='file_dir',
                    help='directory containing static files')
  parser.add_option('-n', '--noprompt', action='store_false',
                    dest='manual_entry',
                    help='disable interactive prompts')
  parser.set_defaults(port=8765,
                      host='maps.google.com',
                      file_dir=FindDefaultFileDir(),
                      manual_entry=True)
  (options, args) = parser.parse_args()

  if not os.path.isfile(os.path.join(options.file_dir, 'index.html')):
    print "Can't find index.html with --file_dir=%s" % options.file_dir
    exit(1)

  if not options.feed_filename and len(args) == 1:
    options.feed_filename = args[0]

  if not options.feed_filename and options.manual_entry:
    options.feed_filename = raw_input('Enter Feed Location: ').strip('"')

  default_key_file = GetDefaultKeyFilePath()
  if not options.key and os.path.isfile(default_key_file):
    options.key = open(default_key_file).read().strip()

  if options.key and os.path.isfile(options.key):
    options.key = open(options.key).read().strip()
    
  # This key is registered to gtfs.schedule.viewer@gmail.com
  if not options.key:
    options.key = 'AIzaSyAZTTRO6RC6LQyKCD3JODhxbClsZl95P9U'

  schedule = transitfeed.Schedule(problem_reporter=transitfeed.ProblemReporter())
  print 'Loading data from feed "%s"...' % options.feed_filename
  print '(this may take a few minutes for larger cities)'
  schedule.Load(options.feed_filename)

  server = StoppableHTTPServer(server_address=('', options.port),
                               RequestHandlerClass=RequestHandlerClass)
  server.key = options.key
  server.schedule = schedule
  server.file_dir = options.file_dir
  server.host = options.host
  server.feed_path = options.feed_filename

  print ("To view, point your browser at http://localhost:%d/" %
         (server.server_port))
  server.serve_forever()


if __name__ == '__main__':
  main()



#!/usr/bin/env python

from setuptools import setup, find_packages


__version__ = '0.0.4'

readme = open("README").read()
changes = open("docs/changes.rst").read()
long_description = readme + "\n\n" + changes


setup(
    name="gtfsni",
    version=__version__,
    author="Jord Flanagan",
    author_email="contact@devopsni.com",
    description="Unofficial Google Transit Feed Generator for Northern Ireland.",
    long_description=long_description,
    url="https://github.com/devopsni/gtfsni",
    download_url="http://pypi.python.org/packages/source/g/gtfsni/gtfsni-%s.tar.gz" % __version__,
    packages = find_packages(),
    package_data = {'gtfsni': [
        'data/translink/*.*',
    ]},
    entry_points = {
        "console_scripts": [
          "gtfs_generate = gtfsni.gtfs_feed_generator:main",
          "gtfs_validate = gtfsni.gtfs_feed_validator:main",
          "gtfs_view = gtfsni.gtfs_schedule_viewer:main",
          "gtfs_to_kml = gtfsni.kmlwriter:main",
          "generate_metro_stops = gtfsni.generate_metro_stops:main",
        ]
    }
)




#!/usr/bin/python2.5
#
# Copyright 2008 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module for writing GTFS feeds out into Google Earth KML format.

For usage information run kmlwriter.py --help

If no output filename is specified, the output file will be given the same
name as the feed file (with ".kml" appended) and will be placed in the same
directory as the input feed.

The resulting KML file has a folder hierarchy which looks like this:

    - Stops
      * stop1
      * stop2
    - Routes
      - route1
        - Shapes
          * shape1
          * shape2
        - Patterns
          - pattern1
          - pattern2
        - Trips
          * trip1
          * trip2
    - Shapes
      * shape1
      - Shape Points
        * shape_point1
        * shape_point2
      * shape2
      - Shape Points
        * shape_point1
        * shape_point2

where the hyphens represent folders and the asteriks represent placemarks.

In a trip, a vehicle visits stops in a certain sequence. Such a sequence of
stops is called a pattern. A pattern is represented by a linestring connecting
the stops. The "Shapes" subfolder of a route folder contains placemarks for
each shape used by a trip in the route. The "Patterns" subfolder contains a
placemark for each unique pattern used by a trip in the route. The "Trips"
subfolder contains a placemark for each trip in the route.

Since there can be many trips and trips for the same route are usually similar,
they are not exported unless the --showtrips option is used. There is also
another option --splitroutes that groups the routes by vehicle type resulting
in a folder hierarchy which looks like this at the top level:

    - Stops
    - Routes - Bus
    - Routes - Tram
    - Routes - Rail
    - Shapes
"""

try:
  import xml.etree.ElementTree as ET  # python 2.5
except ImportError, e:
  import elementtree.ElementTree as ET  # older pythons
import extensions.googletransit as googletransit
import optparse
import os.path
import sys
import transitfeed
from transitfeed import util


class KMLWriter(object):
  """This class knows how to write out a transit feed as KML.

  Sample usage:
    KMLWriter().Write(<transitfeed.Schedule object>, <output filename>)

  Attributes:
    show_trips: True if the individual trips should be included in the routes.
    show_stop_hierarhcy: True if station-stop hierarchy details should be
      included.
    split_routes: True if the routes should be split by type.
    shape_points: True if individual shape points should be plotted.
  """

  def __init__(self):
    """Initialise."""
    self.show_trips = False
    self.show_stop_hierarchy = False
    self.split_routes = False
    self.shape_points = False
    self.altitude_per_sec = 0.0
    self.date_filter = None

  def _SetIndentation(self, elem, level=0):
    """Indented the ElementTree DOM.

    This is the recommended way to cause an ElementTree DOM to be
    prettyprinted on output, as per: http://effbot.org/zone/element-lib.htm

    Run this on the root element before outputting the tree.

    Args:
      elem: The element to start indenting from, usually the document root.
      level: Current indentation level for recursion.
    """
    i = "\n" + level*"  "
    if len(elem):
      if not elem.text or not elem.text.strip():
        elem.text = i + "  "
      for elem in elem:
        self._SetIndentation(elem, level+1)
      if not elem.tail or not elem.tail.strip():
        elem.tail = i
    else:
      if level and (not elem.tail or not elem.tail.strip()):
        elem.tail = i

  def _CreateFolder(self, parent, name, visible=True, description=None):
    """Create a KML Folder element.

    Args:
      parent: The parent ElementTree.Element instance.
      name: The folder name as a string.
      visible: Whether the folder is initially visible or not.
      description: A description string or None.

    Returns:
      The folder ElementTree.Element instance.
    """
    folder = ET.SubElement(parent, 'Folder')
    name_tag = ET.SubElement(folder, 'name')
    name_tag.text = name
    if description is not None:
      desc_tag = ET.SubElement(folder, 'description')
      desc_tag.text = description
    if not visible:
      visibility = ET.SubElement(folder, 'visibility')
      visibility.text = '0'
    return folder

  def _CreateStyleForRoute(self, doc, route):
    """Create a KML Style element for the route.

    The style sets the line colour if the route colour is specified. The
    line thickness is set depending on the vehicle type.

    Args:
      doc: The KML Document ElementTree.Element instance.
      route: The transitfeed.Route to create the style for.

    Returns:
      The id of the style as a string.
    """
    style_id = 'route_%s' % route.route_id
    style = ET.SubElement(doc, 'Style', {'id': style_id})
    linestyle = ET.SubElement(style, 'LineStyle')
    width = ET.SubElement(linestyle, 'width')
    type_to_width = {0: '3',  # Tram
                     1: '3',  # Subway
                     2: '5',  # Rail
                     3: '1'}  # Bus
    width.text = type_to_width.get(route.route_type, '1')
    if route.route_color:
      color = ET.SubElement(linestyle, 'color')
      red = route.route_color[0:2].lower()
      green = route.route_color[2:4].lower()
      blue = route.route_color[4:6].lower()
      color.text = 'ff%s%s%s' % (blue, green, red)
    return style_id

  def _CreatePlacemark(self, parent, name, style_id=None, visible=True,
                       description=None):
    """Create a KML Placemark element.

    Args:
      parent: The parent ElementTree.Element instance.
      name: The placemark name as a string.
      style_id: If not None, the id of a style to use for the placemark.
      visible: Whether the placemark is initially visible or not.
      description: A description string or None.

    Returns:
      The placemark ElementTree.Element instance.
    """
    placemark = ET.SubElement(parent, 'Placemark')
    placemark_name = ET.SubElement(placemark, 'name')
    placemark_name.text = name
    if description is not None:
      desc_tag = ET.SubElement(placemark, 'description')
      desc_tag.text = description
    if style_id is not None:
      styleurl = ET.SubElement(placemark, 'styleUrl')
      styleurl.text = '#%s' % style_id
    if not visible:
      visibility = ET.SubElement(placemark, 'visibility')
      visibility.text = '0'
    return placemark

  def _CreateLineString(self, parent, coordinate_list):
    """Create a KML LineString element.

    The points of the string are given in coordinate_list. Every element of
    coordinate_list should be one of a tuple (longitude, latitude) or a tuple
    (longitude, latitude, altitude).

    Args:
      parent: The parent ElementTree.Element instance.
      coordinate_list: The list of coordinates.

    Returns:
      The LineString ElementTree.Element instance or None if coordinate_list is
      empty.
    """
    if not coordinate_list:
      return None
    linestring = ET.SubElement(parent, 'LineString')
    tessellate = ET.SubElement(linestring, 'tessellate')
    tessellate.text = '1'
    if len(coordinate_list[0]) == 3:
      altitude_mode = ET.SubElement(linestring, 'altitudeMode')
      altitude_mode.text = 'absolute'
    coordinates = ET.SubElement(linestring, 'coordinates')
    if len(coordinate_list[0]) == 3:
      coordinate_str_list = ['%f,%f,%f' % t for t in coordinate_list]
    else:
      coordinate_str_list = ['%f,%f' % t for t in coordinate_list]
    coordinates.text = ' '.join(coordinate_str_list)
    return linestring

  def _CreateLineStringForShape(self, parent, shape):
    """Create a KML LineString using coordinates from a shape.

    Args:
      parent: The parent ElementTree.Element instance.
      shape: The transitfeed.Shape instance.

    Returns:
      The LineString ElementTree.Element instance or None if coordinate_list is
      empty.
    """
    coordinate_list = [(longitude, latitude) for
                       (latitude, longitude, distance) in shape.points]
    return self._CreateLineString(parent, coordinate_list)

  def _CreateStopsFolder(self, schedule, doc):
    """Create a KML Folder containing placemarks for each stop in the schedule.

    If there are no stops in the schedule then no folder is created.

    Args:
      schedule: The transitfeed.Schedule instance.
      doc: The KML Document ElementTree.Element instance.

    Returns:
      The Folder ElementTree.Element instance or None if there are no stops.
    """
    if not schedule.GetStopList():
      return None
    stop_folder = self._CreateFolder(doc, 'Stops')
    stop_folder_selection = self._StopFolderSelectionMethod(stop_folder)
    stop_style_selection = self._StopStyleSelectionMethod(doc)
    stops = list(schedule.GetStopList())
    stops.sort(key=lambda x: x.stop_name)
    for stop in stops:
      (folder, pathway_folder) = stop_folder_selection(stop)
      (style_id, pathway_style_id) = stop_style_selection(stop)
      self._CreateStopPlacemark(folder, stop, style_id)
      if (self.show_stop_hierarchy and
          stop.location_type != transitfeed.Stop.LOCATION_TYPE_STATION and
          stop.parent_station and stop.parent_station in schedule.stops):
        placemark = self._CreatePlacemark(
            pathway_folder, stop.stop_name, pathway_style_id)
        parent_station = schedule.stops[stop.parent_station]
        coordinates = [(stop.stop_lon, stop.stop_lat),
                       (parent_station.stop_lon, parent_station.stop_lat)]
        self._CreateLineString(placemark, coordinates)
    return stop_folder

  def _StopFolderSelectionMethod(self, stop_folder):
    """Create a method to determine which KML folder a stop should go in.

    Args:
      stop_folder: the parent folder element for all stops.

    Returns:
      A function that should accept a Stop argument and return a tuple of
      (stop KML folder, pathways KML folder).

    Given a Stop, we need to determine which folder the stop should go in.  In
    the most basic case, that's the root Stops folder.  However, if
    show_stop_hierarchy is enabled, we put a stop in a separate sub-folder
    depending on if the stop is a station, a platform, an entrance, or just a
    plain-old stand-alone stop.  This method returns a function that is used
    to pick which folder a stop stop should go in.  It also optionally returns
    a folder where any line-string connections associated with a stop (eg. to
    show the pathway between an entrance and a station) should be added.
    """
    if not self.show_stop_hierarchy:
      return lambda stop: (stop_folder, None)

    # Create the various sub-folders for showing the stop hierarchy
    station_folder = self._CreateFolder(stop_folder, 'Stations')
    platform_folder = self._CreateFolder(stop_folder, 'Platforms')
    platform_connections = self._CreateFolder(platform_folder, 'Connections')
    entrance_folder = self._CreateFolder(stop_folder, 'Entrances')
    entrance_connections = self._CreateFolder(entrance_folder, 'Connections')
    standalone_folder = self._CreateFolder(stop_folder, 'Stand-Alone')

    def FolderSelectionMethod(stop):
      if stop.location_type == transitfeed.Stop.LOCATION_TYPE_STATION:
        return (station_folder, None)
      elif stop.location_type == googletransit.Stop.LOCATION_TYPE_ENTRANCE:
        return (entrance_folder, entrance_connections)
      elif stop.parent_station:
        return (platform_folder, platform_connections)
      return (standalone_folder, None)

    return FolderSelectionMethod

  def _StopStyleSelectionMethod(self, doc):
    """Create a method to determine which style to apply to a stop placemark.

    Args:
      doc: the KML document.

    Returns:
      A function that should accept a Stop argument and return a tuple of
      (stop placemark style id, pathway placemark style id).  Either style id
      can be None, indicating no style should be set.

    Given a Stop, we need to determine what KML style to apply to the stops'
    placemark.  In the most basic case, no styling is applied.  However, if
    show_stop_hierarchy is enabled, we style each type of stop differently
    depending on if the stop is a station, platform, entrance, etc.  This method
    returns a function that is used to pick which style id should be associated
    with a stop placemark, or None if no style should be applied.  It also
    optionally returns a style id to associate with any line-string connections
    associated with a stop (eg. to show the pathway between an entrance and a
    station).
    """
    if not self.show_stop_hierarchy:
      return lambda stop: (None, None)

    # Create the various styles for showing the stop hierarchy
    self._CreateStyle(
        doc, 'stop_entrance', {'IconStyle': {'color': 'ff0000ff'}})
    self._CreateStyle(
        doc,
        'entrance_connection',
        {'LineStyle': {'color': 'ff0000ff', 'width': '2'}})
    self._CreateStyle(
        doc, 'stop_platform', {'IconStyle': {'color': 'ffff0000'}})
    self._CreateStyle(
        doc,
        'platform_connection',
        {'LineStyle': {'color': 'ffff0000', 'width': '2'}})
    self._CreateStyle(
        doc, 'stop_standalone', {'IconStyle': {'color': 'ff00ff00'}})

    def StyleSelectionMethod(stop):
      if stop.location_type == transitfeed.Stop.LOCATION_TYPE_STATION:
        return ('stop_station', None)
      elif stop.location_type == googletransit.Stop.LOCATION_TYPE_ENTRANCE:
        return ('stop_entrance', 'entrance_connection')
      elif stop.parent_station:
        return ('stop_platform', 'platform_connection')
      return ('stop_standalone', None)

    return StyleSelectionMethod

  def _CreateStyle(self, doc, style_id, style_dict):
    """Helper method to create a <Style/> element in a KML document.

    Args:
      doc: the parent KML document.
      style_id: the style id to include for the <Style/> element.
      style_dict: a dict of sub-elements and values to add to the <Style/>.

    Returns:
      The newly created <Style/> element.

    Each key of the style_dict argument is used to create a sub-element of the
    parent <Style/> element.  If the value associated with that key is a string,
    then it will be used to set the text value of the sub-element.  If the value
    is another dict, it will be used to recursively construct a sub-sub-element
    with the same semantics.
    """

    def CreateElements(current_element, current_dict):
      for (key, value) in current_dict.iteritems():
        element = ET.SubElement(current_element, key)
        if isinstance(value,dict):
          CreateElements(element, value)
        else:
          element.text = value

    style = ET.SubElement(doc, 'Style', {'id': style_id})
    CreateElements(style, style_dict)
    return style

  def _CreateStopPlacemark(self, stop_folder, stop, style_id):
    """Creates a new stop <Placemark/> element.

    Args:
      stop_folder: the KML folder the placemark will be added to.
      stop: the actual Stop to create a placemark for.
      style_id: optional argument indicating a style id to add to the placemark.
    """
    desc_items = []
    desc_items.append("Stop id: %s" % stop.stop_id)
    if stop.stop_desc:
      desc_items.append(stop.stop_desc)
    if stop.stop_url:
      desc_items.append('Stop info page: <a href="%s">%s</a>' % (
          stop.stop_url, stop.stop_url))
    description = '<br/>'.join(desc_items) or None
    placemark = self._CreatePlacemark(stop_folder, stop.stop_name,
                                        description=description,
                                        style_id=style_id)
    point = ET.SubElement(placemark, 'Point')
    coordinates = ET.SubElement(point, 'coordinates')
    coordinates.text = '%.6f,%.6f' % (stop.stop_lon, stop.stop_lat)

  def _CreateRoutePatternsFolder(self, parent, route,
                                   style_id=None, visible=True):
    """Create a KML Folder containing placemarks for each pattern in the route.

    A pattern is a sequence of stops used by one of the trips in the route.

    If there are not patterns for the route then no folder is created and None
    is returned.

    Args:
      parent: The parent ElementTree.Element instance.
      route: The transitfeed.Route instance.
      style_id: The id of a style to use if not None.
      visible: Whether the folder is initially visible or not.

    Returns:
      The Folder ElementTree.Element instance or None if there are no patterns.
    """
    pattern_id_to_trips = route.GetPatternIdTripDict()
    if not pattern_id_to_trips:
      return None

    # sort by number of trips using the pattern
    pattern_trips = pattern_id_to_trips.values()
    pattern_trips.sort(lambda a, b: cmp(len(b), len(a)))

    folder = self._CreateFolder(parent, 'Patterns', visible)
    for n, trips in enumerate(pattern_trips):
      trip_ids = [trip.trip_id for trip in trips]
      name = 'Pattern %d (trips: %d)' % (n+1, len(trips))
      description = 'Trips using this pattern (%d in total): %s' % (
          len(trips), ', '.join(trip_ids))
      placemark = self._CreatePlacemark(folder, name, style_id, visible,
                                        description)
      coordinates = [(stop.stop_lon, stop.stop_lat)
                     for stop in trips[0].GetPattern()]
      self._CreateLineString(placemark, coordinates)
    return folder

  def _CreateRouteShapesFolder(self, schedule, parent, route,
                               style_id=None, visible=True):
    """Create a KML Folder for the shapes of a route.

    The folder contains a placemark for each shape referenced by a trip in the
    route. If there are no such shapes, no folder is created and None is
    returned.

    Args:
      schedule: The transitfeed.Schedule instance.
      parent: The parent ElementTree.Element instance.
      route: The transitfeed.Route instance.
      style_id: The id of a style to use if not None.
      visible: Whether the placemark is initially visible or not.

    Returns:
      The Folder ElementTree.Element instance or None.
    """
    shape_id_to_trips = {}
    for trip in route.trips:
      if trip.shape_id:
        shape_id_to_trips.setdefault(trip.shape_id, []).append(trip)
    if not shape_id_to_trips:
      return None

    # sort by the number of trips using the shape
    shape_id_to_trips_items = shape_id_to_trips.items()
    shape_id_to_trips_items.sort(lambda a, b: cmp(len(b[1]), len(a[1])))

    folder = self._CreateFolder(parent, 'Shapes', visible)
    for shape_id, trips in shape_id_to_trips_items:
      trip_ids = [trip.trip_id for trip in trips]
      name = '%s (trips: %d)' % (shape_id, len(trips))
      description = 'Trips using this shape (%d in total): %s' % (
          len(trips), ', '.join(trip_ids))
      placemark = self._CreatePlacemark(folder, name, style_id, visible,
                                        description)
      self._CreateLineStringForShape(placemark, schedule.GetShape(shape_id))
    return folder

  def _CreateRouteTripsFolder(self, parent, route, style_id=None, schedule=None):
    """Create a KML Folder containing all the trips in the route.

    The folder contains a placemark for each of these trips. If there are no
    trips in the route, no folder is created and None is returned.

    Args:
      parent: The parent ElementTree.Element instance.
      route: The transitfeed.Route instance.
      style_id: A style id string for the placemarks or None.

    Returns:
      The Folder ElementTree.Element instance or None.
    """
    if not route.trips:
      return None
    trips = list(route.trips)
    trips.sort(key=lambda x: x.trip_id)
    trips_folder = self._CreateFolder(parent, 'Trips', visible=False)
    for trip in trips:
      if (self.date_filter and
          not trip.service_period.IsActiveOn(self.date_filter)):
        continue

      if trip.trip_headsign:
        description = 'Headsign: %s' % trip.trip_headsign
      else:
        description = None

      coordinate_list = []
      for secs, stoptime, tp in trip.GetTimeInterpolatedStops():
        if self.altitude_per_sec > 0:
          coordinate_list.append((stoptime.stop.stop_lon, stoptime.stop.stop_lat,
                                  (secs - 3600 * 4) * self.altitude_per_sec))
        else:
          coordinate_list.append((stoptime.stop.stop_lon,
                                  stoptime.stop.stop_lat))
      placemark = self._CreatePlacemark(trips_folder,
                                        trip.trip_id,
                                        style_id=style_id,
                                        visible=False,
                                        description=description)
      self._CreateLineString(placemark, coordinate_list)
    return trips_folder

  def _CreateRoutesFolder(self, schedule, doc, route_type=None):
    """Create a KML Folder containing routes in a schedule.

    The folder contains a subfolder for each route in the schedule of type
    route_type. If route_type is None, then all routes are selected. Each
    subfolder contains a flattened graph placemark, a route shapes placemark
    and, if show_trips is True, a subfolder containing placemarks for each of
    the trips in the route.

    If there are no routes in the schedule then no folder is created and None
    is returned.

    Args:
      schedule: The transitfeed.Schedule instance.
      doc: The KML Document ElementTree.Element instance.
      route_type: The route type integer or None.

    Returns:
      The Folder ElementTree.Element instance or None.
    """

    def GetRouteName(route):
      """Return a placemark name for the route.

      Args:
        route: The transitfeed.Route instance.

      Returns:
        The name as a string.
      """
      name_parts = []
      if route.route_short_name:
        name_parts.append('<b>%s</b>' % route.route_short_name)
      if route.route_long_name:
        name_parts.append(route.route_long_name)
      return ' - '.join(name_parts) or route.route_id

    def GetRouteDescription(route):
      """Return a placemark description for the route.

      Args:
        route: The transitfeed.Route instance.

      Returns:
        The description as a string.
      """
      desc_items = []
      if route.route_desc:
        desc_items.append(route.route_desc)
      if route.route_url:
        desc_items.append('Route info page: <a href="%s">%s</a>' % (
            route.route_url, route.route_url))
      description = '<br/>'.join(desc_items)
      return description or None

    routes = [route for route in schedule.GetRouteList()
              if route_type is None or route.route_type == route_type]
    if not routes:
      return None
    routes.sort(key=lambda x: GetRouteName(x))

    if route_type is not None:
      route_type_names = {0: 'Tram, Streetcar or Light rail',
                          1: 'Subway or Metro',
                          2: 'Rail',
                          3: 'Bus',
                          4: 'Ferry',
                          5: 'Cable car',
                          6: 'Gondola or suspended cable car',
                          7: 'Funicular'}
      type_name = route_type_names.get(route_type, str(route_type))
      folder_name = 'Routes - %s' % type_name
    else:
      folder_name = 'Routes'
    routes_folder = self._CreateFolder(doc, folder_name, visible=False)

    for route in routes:
      style_id = self._CreateStyleForRoute(doc, route)
      route_folder = self._CreateFolder(routes_folder,
                                        GetRouteName(route),
                                        description=GetRouteDescription(route))
      self._CreateRouteShapesFolder(schedule, route_folder, route,
                                    style_id, False)
      self._CreateRoutePatternsFolder(route_folder, route, style_id, False)
      if self.show_trips:
        self._CreateRouteTripsFolder(route_folder, route, style_id, schedule)
    return routes_folder

  def _CreateShapesFolder(self, schedule, doc):
    """Create a KML Folder containing all the shapes in a schedule.

    The folder contains a placemark for each shape. If there are no shapes in
    the schedule then the folder is not created and None is returned.

    Args:
      schedule: The transitfeed.Schedule instance.
      doc: The KML Document ElementTree.Element instance.

    Returns:
      The Folder ElementTree.Element instance or None.
    """
    if not schedule.GetShapeList():
      return None
    shapes_folder = self._CreateFolder(doc, 'Shapes')
    shapes = list(schedule.GetShapeList())
    shapes.sort(key=lambda x: x.shape_id)
    for shape in shapes:
      placemark = self._CreatePlacemark(shapes_folder, shape.shape_id)
      self._CreateLineStringForShape(placemark, shape)
      if self.shape_points:
        self._CreateShapePointFolder(shapes_folder, shape)
    return shapes_folder

  def _CreateShapePointFolder(self, shapes_folder, shape):
    """Create a KML Folder containing all the shape points in a shape.

    The folder contains placemarks for each shapepoint.

    Args:
      shapes_folder: A KML Shape Folder ElementTree.Element instance
      shape: The shape to plot.

    Returns:
      The Folder ElementTree.Element instance or None.
    """

    folder_name = shape.shape_id + ' Shape Points'
    folder = self._CreateFolder(shapes_folder, folder_name, visible=False)
    for (index, (lat, lon, dist)) in enumerate(shape.points):
      placemark = self._CreatePlacemark(folder, str(index+1))
      point = ET.SubElement(placemark, 'Point')
      coordinates = ET.SubElement(point, 'coordinates')
      coordinates.text = '%.6f,%.6f' % (lon, lat)
    return folder

  def Write(self, schedule, output_file):
    """Writes out a feed as KML.

    Args:
      schedule: A transitfeed.Schedule object containing the feed to write.
      output_file: The name of the output KML file, or file object to use.
    """
    # Generate the DOM to write
    root = ET.Element('kml')
    root.attrib['xmlns'] = 'http://earth.google.com/kml/2.1'
    doc = ET.SubElement(root, 'Document')
    open_tag = ET.SubElement(doc, 'open')
    open_tag.text = '1'
    self._CreateStopsFolder(schedule, doc)
    if self.split_routes:
      route_types = set()
      for route in schedule.GetRouteList():
        route_types.add(route.route_type)
      route_types = list(route_types)
      route_types.sort()
      for route_type in route_types:
        self._CreateRoutesFolder(schedule, doc, route_type)
    else:
      self._CreateRoutesFolder(schedule, doc)
    self._CreateShapesFolder(schedule, doc)

    # Make sure we pretty-print
    self._SetIndentation(root)

    # Now write the output
    if isinstance(output_file, file):
      output = output_file
    else:
      output = open(output_file, 'w')
    output.write("""<?xml version="1.0" encoding="UTF-8"?>\n""")
    ET.ElementTree(root).write(output, 'utf-8')


def main():
  usage = \
'''%prog [options] <input GTFS.zip> [<output.kml>]

Reads GTFS file or directory <input GTFS.zip> and creates a KML file
<output.kml> that contains the geographical features of the input. If
<output.kml> is omitted a default filename is picked based on
<input GTFS.zip>. By default the KML contains all stops and shapes.

For more information see
http://code.google.com/p/googletransitdatafeed/wiki/KMLWriter
'''

  parser = util.OptionParserLongError(
      usage=usage, version='%prog '+transitfeed.__version__)
  parser.add_option('-t', '--showtrips', action='store_true',
                    dest='show_trips',
                    help='include the individual trips for each route')
  parser.add_option('-a', '--altitude_per_sec', action='store', type='float',
                    dest='altitude_per_sec',
                    help='if greater than 0 trips are drawn with time axis '
                    'set to this many meters high for each second of time')
  parser.add_option('-s', '--splitroutes', action='store_true',
                    dest='split_routes',
                    help='split the routes by type')
  parser.add_option('-d', '--date_filter', action='store', type='string',
                    dest='date_filter',
                    help='Restrict to trips active on date YYYYMMDD')
  parser.add_option('-p', '--display_shape_points', action='store_true',
                    dest='shape_points',
                    help='shows the actual points along shapes')
  parser.add_option('--show_stop_hierarchy', action='store_true',
                    dest='show_stop_hierarchy',
                    help='include station-stop hierarchy info in output')

  parser.set_defaults(altitude_per_sec=1.0)
  options, args = parser.parse_args()

  if len(args) < 1:
    parser.error('You must provide the path of an input GTFS file.')

  if args[0] == 'IWantMyCrash':
    raise Exception('For testCrashHandler')

  input_path = args[0]
  if len(args) >= 2:
    output_path = args[1]
  else:
    path = os.path.normpath(input_path)
    (feed_dir, feed) = os.path.split(path)
    if '.' in feed:
      feed = feed.rsplit('.', 1)[0]  # strip extension
    output_filename = '%s.kml' % feed
    output_path = os.path.join(feed_dir, output_filename)

  loader = transitfeed.Loader(input_path,
                              problems=transitfeed.ProblemReporter())
  feed = loader.Load()
  print "Writing %s" % output_path
  writer = KMLWriter()
  writer.show_trips = options.show_trips
  writer.altitude_per_sec = options.altitude_per_sec
  writer.split_routes = options.split_routes
  writer.date_filter = options.date_filter
  writer.shape_points = options.shape_points
  writer.show_stop_hierarchy = options.show_stop_hierarchy
  writer.Write(feed, output_path)


if __name__ == '__main__':
  util.RunWithCrashHandler(main)




import os
import csv
from itertools import groupby
import time
import shutil

import requests

from gtfsni import get_pkg_data, get_app_data
from gtfsni import utils
from gtfsni.generate_metro_stops import main as generate_metro_stops

ENDPOINT = 'https://api.scraperwiki.com/api/1.0/datastore/sqlite'
TRIPDATAJOINER = '|'
TRIPDATAPARTJOINER = '#'
RAW_TIMETABLE_SCHEMA = (
    'agency_id', 'operator_id', 'service_id',
    'route_id', 'route_url', 'route_type', 'route_long_name',
    'route_short_name', 'route_begin', 'route_end', 'route_direction',
    'stop_name', 'stop_sequence', 'schedule',
)
GTFS_SCHEMA_MAP = {
    'stops.txt': [(
        'stop_lat', 'stop_lon', 'stop_name', 'road', 'direction', 'geohash',
    ),(
        'stop_id', 'stop_code', 'stop_name', 'stop_desc', 'stop_lat', 'stop_lon',
        'zone_id', 'stop_url', 'location_type', 'parent_station', 'stop_timezone',
        'road', 'direction', 'geohash',
    )],
    'routes.txt': [(
        'agency_id', 'route_id', 'route_type', 'route_url',
        'route_short_name', 'route_long_name',
    ),(
        'agency_id', 'route_id', 'route_type', 'route_url',
        'route_short_name', 'route_long_name', 'route_desc', 
        'route_color', 'route_text_color',
    )],
    'trips.txt': [
        '*'
    ,(
        'route_id', 'service_id', 'trip_id', 'trip_headsign', 'trip_short_name',
        'direction_id', 'block_id', 'shape_id', 'timeframe_id', 'trip_sequence',
    )],
    'calendar.txt': [
        None
    ,(
        'service_id',
        'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday',
        'start_date', 'end_date',
    )],
    'stop_times.txt': [
        '*'
    ,(
        'trip_id', 'arrival_time', 'departure_time', 'stop_id', 'stop_sequence',
        'stop_headsign', 'pickup_type', 'drop_off_type', 'shape_dist_travelled',
    )],
}


def make_id(*args):
    return utils.slugify(' '.join(str(arg) for arg in args))

def make_route_id(*args):
    return make_id(*args).replace('-', '').upper()

def make_trip_id(routeid, timeframeid, tripno):
    return make_id(routeid, timeframeid, '%03d' % int(tripno))

def CSVReader(schema):
    sql = "SELECT %s FROM `swdata` WHERE operator_id='METRO'"
    if schema == '*':
        schema = RAW_TIMETABLE_SCHEMA
        query = ', '.join(schema)
    else:
        query = 'DISTINCT ' + ', '.join(schema)
    sql %= query
    params = {
        'name': 'translink_ni_timetables',
        'format': 'csv',
        'query': sql,
    }
    response = requests.get(ENDPOINT, params=params)
    reader = csv.DictReader(response.text.splitlines(), schema)
    reader.next()
    return sorted(reader, key=lambda row: tuple(row.values()))

class CSVWriter(object):

    def __init__(self, filename, schema):
        self.filename = get_app_data('gtfs/' + filename)
        self.schema = schema
        self._fd = open(self.filename, 'wb')
        self._writer = csv.DictWriter(self._fd, schema)
        self._write_header()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def _write_header(self):
        self._writer.writerow(dict(zip(self.schema, self.schema)))

    def writerow(self, **kw):
        row = dict((field, None) for field in self.schema)
        row.update(kw)
        self._writer.writerow(row)
        return row

    def close(self):
        try:
            self._fd.close()
        except:
            pass

def get_schemas(fname):
    schema_in, schema_out = GTFS_SCHEMA_MAP[fname]
    if schema_in and schema_out and schema_in != '*':
        assert set(schema_in) <= set(schema_out), ', '.join(schema_out)
    return schema_in, schema_out

def write_agency():
    src = get_pkg_data('translink/agency.csv')
    dest = get_app_data('gtfs/agency.txt')
    shutil.copyfile(src, dest)

def write_routes():
    fname = 'routes.txt'
    schema_in, schema_out = get_schemas(fname)
    reader = CSVReader(schema_in)
    with CSVWriter(fname, schema_out) as writer:
        for row in reader:
            writer.writerow(**row)

def write_stops():
    fname = 'stops.txt'
    schema_in, schema_out = get_schemas(fname)
    metro_stops = get_app_data('translink/metro_stops.csv')
    if not os.path.exists(metro_stops):
        raise Exception("stop coordinate file not found - %s" % metro_stops)
    direction2code = utils.direction2code
    with open(metro_stops) as fd:
        reader = csv.DictReader(fd, schema_in)
        reader.next()
        with CSVWriter(fname, schema_out) as writer:
            for row in reader:
                stop, direction = row['stop_name'], row['direction']
                lat, lon = row['stop_lat'], row['stop_lon']
                if not lat or not lon:
                    continue
                id = make_id('translinkni', stop, direction2code[direction])
                writer.writerow(
                    stop_id=id,
                    stop_name=row['stop_name'],
                    stop_lat=row['stop_lat'],
                    stop_lon=row['stop_lon'],
                    stop_desc=row['road'],
                )

def write_trips_and_calendar():
    alltripids = []
    tripsfile = 'trips.txt'
    schema_in, trips_schema = get_schemas(tripsfile)
    reader = CSVReader(schema_in)
    sortkey = lambda X: (X['agency_id'], X['route_id'])
    data = sorted(reader, key=sortkey)
    services = {}
    with CSVWriter(tripsfile, trips_schema) as writer:
        for key, g in groupby(data, key=sortkey):
            # g contains one row for every stop on the route but the data
            # we want next - trip number and service id - is the same for
            # each row (in the group), so we just use the first row and continue.
            # the time schedule string should never be jagged for this to work
            # ie. each stop has a value for each trip, even if it is '...'
            agencyid, routeid = key
            row = g.next()
            direction = row['route_direction']
            route_end = row['route_end']
            tripdata = (tuple(t.split(TRIPDATAPARTJOINER)[:3]) for t in row['schedule'].split(TRIPDATAJOINER) if t)
            service_id = row['service_id']
            # inbound and outbound routes have the same service id
            service_calendar = services.setdefault(service_id, {
                'service_id': service_id,
                'monday': 0,
                'tuesday': 0,
                'wednesday': 0,
                'thursday': 0,
                'friday': 0,
                'saturday': 0,
                'sunday': 0,
                'start_date':'20110101',
                'end_date':'20991231',
            })
            for tripno, srvno, timeframe in tripdata:
                #use numeric timeframe code in trip id
                #tfname, tfweight = get_time_period_name_and_weight(timeframe)
                tripno = int(tripno)
                headsign = '%s - %s' % (srvno, route_end)
                for idx, daycode in utils.split_timeframe(timeframe):
                    idx = str(idx)
                    tripid = make_trip_id(routeid, idx, tripno)
                    writer.writerow(
                        route_id=routeid, service_id=service_id, trip_id=tripid,
                        trip_headsign=headsign, trip_short_name=headsign,
                        direction_id=direction, timeframe_id=daycode,
                        trip_sequence=tripno
                    )
                    alltripids.append(tripid)
                    dayname = utils.daycode_to_name[daycode]
                    service_calendar[dayname.lower()] = 1
    calendarfile = 'calendar.txt'
    _, calendar_schema = get_schemas(calendarfile)
    # write 'calendar.txt'
    with CSVWriter(calendarfile, calendar_schema) as writer:
        for k in sorted(services.keys()):
            writer.writerow(**services[k])
    #assert trip_ids are unique
    assert len(alltripids) == len(set(alltripids))

def write_stop_times():
    fname = 'stop_times.txt'
    schema_in, schema_out = get_schemas(fname)
    reader = CSVReader(schema_in)
    with CSVWriter(fname, schema_out) as writer:
        for row in reader:
            agencyid = row['agency_id']
            routeid = row['route_id']
            stopname = row['stop_name']
            direction = row['route_direction']
            stopid = make_id(agencyid, stopname, direction)
            stopno = row['stop_sequence']
            headsign = '%s %s' % (row['route_short_name'], row['route_end'])
            stoptime_data = (
                tuple(
                    t.split(TRIPDATAPARTJOINER)
                ) for t in row['schedule'].split(TRIPDATAJOINER) if t
            )
            for tripno, srvno, timeframe, time in stoptime_data:
                #ignore any special instruction code for the minute
                if time == '...':
                    #time = None
                    continue
                else:
                    time = '%s:%s:00' % (time[:2], time[2:4])
                for idx, daycode in utils.split_timeframe(timeframe):
                    idx = str(idx)
                    tripid = make_trip_id(routeid, idx, tripno)
                    t = (tripid, time, time, stopid, stopno, stopname, None, None, None)
                    writer.writerow(
                        trip_id=tripid, stop_id=stopid, stop_sequence=stopno,
                        arrival_time=time, departure_time=time,
                        stop_headsign=headsign,
                    )

def compress():
    timestamp = time.strftime('%Y_%m_%d_%H%M%S', time.localtime())
    archive = 'google_transit_translinkni_%s.zip' % timestamp
    cxt = {
        'archive': archive,
        'src': get_app_data('gtfs/'),
        'dest': get_app_data('dist/' + archive),
        'destroot': get_app_data('dist/'),
        'symlink': 'google_transit.zip',
    }
    cmd = (
        'cd %(src)s && zip %(archive)s *.txt && mv %(archive)s %(dest)s'
        ' && cd %(destroot)s && rm -f %(symlink)s'
        ' && ln -s %(archive)s %(symlink)s'
    )
    cmd %= cxt
    ret = os.system(cmd)
    if ret:
        raise OSError(cmd)
    print 'created archive: %s' % cxt['dest']

def main():
    generate_metro_stops()
    write_stops()
    write_routes()
    write_trips_and_calendar()
    write_stop_times()
    write_agency()
    compress()


